import time
import unittest
import warnings

from vika import Vika

from . import TOKEN, DOMAIN, SPACE_ID, DATASHEET_ID


class TestUpdateRecords(unittest.TestCase):
    def setUp(self):
        warnings.simplefilter("ignore", ResourceWarning)
        apitable = Vika(TOKEN)
        apitable.set_api_base(DOMAIN)
        self.dst = apitable.space(SPACE_ID).datasheet(DATASHEET_ID)

    def test_record_update(self):
        # Update a single field
        record = self.dst.records.get(title="apitable")
        record.title = "apitable_py"
        self.assertEqual(record.title, "apitable_py")
        time.sleep(1 / 5)

        # Update multiple fields
        record = self.dst.records.get(title="apitable_py")
        r = record.update({"title": "apitable_api", "comment": "apitable rest api"})
        self.assertEqual(r.title, "apitable_api")
        self.assertEqual(r.comment, "apitable rest api")
        time.sleep(1 / 5)

        # Update multiple records
        self.dst.records.filter(title="apitable_api").update(title="apitable_widget")
        record = self.dst.records.get(title="apitable_widget")
        self.assertEqual(record.title, "apitable_widget")
        time.sleep(1 / 5)

    def test_bulk_update(self):
        records = self.dst.records.bulk_create(
            [
                {"title": "apitable_py"},
                {"title": "apitable_api"},
                {"title": "apitable_widget"},
            ]
        )
        update_data = [
            {"recordId": rec._id, "fields": {"title": "new" + rec.title}}
            for rec in records
        ]
        updated_records = self.dst.records.bulk_update(update_data)
        self.assertEqual(updated_records[0].title, "new" + records[0].title)
        self.dst.delete_records(records)

    def test_update_or_create(self):
        record, created = self.dst.records.update_or_create(
            title="apitable_api",
            defaults={"title": "apitable_widget", "comment": "apitable widget sdk"},
        )
        self.assertTrue(created)
        self.assertEqual(record.comment, "apitable widget sdk")
        time.sleep(1 / 5)
        # Create a record
        record, created = self.dst.records.update_or_create(
            title="test_record", defaults={"comment": "apitable rest api"}
        )
        self.assertTrue(created)
        self.assertEqual(record.comment, "apitable rest api")
        self.assertEqual(record.title, "test_record")
        self.dst.records.filter(title="test_record").delete()
        self.dst.records.filter(title="apitable_widget").delete()


if __name__ == "__main__":
    unittest.main()



from typing import List

from vika.const import MAX_WRITE_RECORDS_PRE_REQ
from vika.datasheet.record import Record
from vika.exceptions import RecordDoesNotExist
from vika.types import RawRecord
from vika.utils import chunks, trans_key


class QuerySet:

    def __init__(self, dst, records: List[RawRecord]):
        self._dst = dst
        self._records = records[:]

    def __len__(self):
        return len(self._records)

    def __iter__(self):
        for record in iter(self._records):
            record = Record(self._dst, record)
            yield record

    def __str__(self):
        return f"(QuerySet: {self.count()} records)"

    __repr__ = __str__

    def __getitem__(self, index):
        return Record(self._dst, self._records[index])

    def chunks(self):
        """
        Split the current QuerySet into multiple QuerySets with a maximum number of records of 10
        @return: iter[QuerySet] iteratively returns a sub-QuerySet with a maximum number of records of 10
        """
        for chunk in chunks(self._records, MAX_WRITE_RECORDS_PRE_REQ):
            yield QuerySet(self._dst, chunk)

    def last(self):
        """
        Get the last record
        @return: Record
        """
        return Record(self._dst, self._records[-1])

    def first(self):
        """
        Get the first record
        @return: Record
        """
        return Record(self._dst, self._records[0])

    def delete(self, delete_all=False) -> bool:
        """
        delete Delete all records in the current record set in batches, delete in batches of 10 records. If the number of recordsets is greater than 10, an error will be reported
        @return:
        """
        if not delete_all and self.count() > MAX_WRITE_RECORDS_PRE_REQ:
            raise Exception('You cannot operate more than 10 records in batches, please use the chunks method to operate in batches.')
        return self._dst.delete_records([rec.id for rec in self._records])

    def clone(self):
        """
        Returns a cloned copy of the current recordset
        @return: QuerySet
        """
        return QuerySet(self._dst, self._records[:])

    def update(self, **kwargs) -> bool:
        """
        dst.records.filter(title=None).update(status="Pending")
        """
        if self.count() > MAX_WRITE_RECORDS_PRE_REQ:
            raise Exception('You cannot operate more than 10 records in batches, please use the chunks method to operate in batches')
        patch_update_records_data = []
        for record in iter(self._records):
            data = {"recordId": record.id, "fields": kwargs}
            patch_update_records_data.append(data)
        self._records = self._dst.update_records(patch_update_records_data)
        return True

    def count(self):
        """
        Returns the total number of records in the current recordset
        @return: int
        """
        return len(self)

    def get(self, **kwargs):
        if kwargs:
            return self.filter(**kwargs).first()
        if self._records:
            return Record(self._dst, self._records[0])
        raise RecordDoesNotExist()

    def filter(self, *args, **kwargs):
        """
        songs = dst_songs.filter(artist="faye wong")
        songs = dst_songs.filter('{artist}="faye wong" OR {artist}="someone"')
        for song in songs:
            print(song.title)
        """
        if args and kwargs:
            #
            print()
        # if filter_func_or_key:
        #     if isinstance(filter_func_or_key, FunctionType):
        #         if kwargs:
        #             raise Exception("function filter can't be used with key-value filter")
        kwargs = {
            trans_key(self._dst.field_key_map, k): v
            for k, v in kwargs.items()
        }

        def filter_record(record) -> bool:
            return all([
                record.id == v
                if k in ["recordId", "_id"] else record.data.get(k) == v
                for k, v in kwargs.items()
            ])

        found_records = list(filter(filter_record, self._records))
        return QuerySet(self._dst, found_records)



from typing import List

from vika.types import MetaView


class ViewManager:

    def __init__(self, dst: 'Datasheet'):
        self.dst = dst

    def all(self) -> List[MetaView]:
        """
        Query all views of the current datasheet
        @return: List[MetaView]
        """
        return self.dst.get_views()



import unittest

from apitable.utils import query_parse


class TestUtils(unittest.TestCase):

    def test_query_parse(self):
        field_key_map = {}
        self.assertEqual(query_parse(field_key_map, a=1), "{a}=1")
        self.assertEqual(query_parse(field_key_map, a=1, b=2),
                         "AND({a}=1,{b}=2)")
        self.assertEqual(query_parse(field_key_map, a="1", b=2),
                         'AND({a}="1",{b}=2)')
        self.assertEqual(query_parse(field_key_map, a=1, b=['B', 'C', 'D']),
                         'AND({a}=1,{b}="B, C, D")')
        self.assertEqual(query_parse(field_key_map, a=1, b=True),
                         "AND({a}=1,{b}=TRUE())")
        self.assertEqual(query_parse(field_key_map, a=1, b=False),
                         "AND({a}=1,{b}=FALSE())")
        self.assertEqual(query_parse(field_key_map, a=1, b=None),
                         "AND({a}=1,{b}=BLANK())")


if __name__ == "__main__":
    unittest.main()



import unittest
import warnings

from vika import Vika

from . import TOKEN, DOMAIN


class TestGetSpaces(unittest.TestCase):
    def setUp(self):
        warnings.simplefilter("ignore", ResourceWarning)
        apitable = Vika(TOKEN)
        apitable.set_api_base(DOMAIN)
        self.apitable = apitable

    def test_spaces_all(self):
        spaces = self.apitable.spaces.all()
        self.assertIsInstance(spaces, list)
        first_space = spaces[0]
        self.assertIsNotNone(first_space.name)
        self.assertIsInstance(first_space.name, str)


if __name__ == "__main__":
    unittest.main()



from typing import Dict, Any, List, Optional

from vika.types.embedlink import EmbedLinkThemeEnum, EmbedLinkPayload
from vika.types.node import NodeListItem, NodeDetail
from vika.types.space import SpaceListItem
from vika.types.record import RawRecord
from vika.types.view import MetaView
from vika.types.field import MetaField
from pydantic import BaseModel


class ResponseBase(BaseModel):
    """
    Unified response body format returned by REST API
    """
    code: int
    success: bool
    message: str
    data: Optional[Any]


Records = List[RawRecord]


class RecordsData(BaseModel):
    records: Records


class GETResponseBase(BaseModel):
    total: int
    pageNum: int
    pageSize: int


class GETRecordResponseData(GETResponseBase, RecordsData):
    pass


class GETRecordResponse(ResponseBase):
    data: GETRecordResponseData


class PatchRecordResponse(ResponseBase):
    data: RecordsData


class PostRecordResponse(ResponseBase):
    data: RecordsData


class DeleteRecordResponse(ResponseBase):
    pass


class UploadFileResponse(ResponseBase):
    data: Dict[str, Any]


# meta field
class GETMetaFieldResponseData(BaseModel):
    items: List[MetaField]

    class Config:
        # https://github.com/samuelcolvin/pydantic/issues/1250
        fields = {
            "items": "fields",  # The data returned by the server is fields, which are reserved for pydantic.
        }


class GETMetaFieldResponse(ResponseBase):
    data: GETMetaFieldResponseData


class PostMetaFieldResponseData(BaseModel):
    id: str
    name: str


class PostMetaFieldResponse(ResponseBase):
    data: PostMetaFieldResponseData


class DeleteFieldResponse(ResponseBase):
    pass


# meta view
class GETMetaViewResponseData(BaseModel):
    views: List[MetaView]


class GETMetaViewResponse(ResponseBase):
    data: GETMetaViewResponseData


# space
class GETSpaceListResponseData(BaseModel):
    spaces: List[SpaceListItem]


class GETSpaceListResponse(ResponseBase):
    data: GETSpaceListResponseData


# node
class GETNodeListResponseData(BaseModel):
    nodes: List[NodeListItem]


class GETNodeListResponse(ResponseBase):
    data: GETNodeListResponseData


class GETNodeDetailResponse(ResponseBase):
    data: NodeDetail


class PostDatasheetMetaResponseData(BaseModel):
    id: str
    createdAt: int
    items: List[PostMetaFieldResponseData]

    class Config:
        # https://github.com/samuelcolvin/pydantic/issues/1250
        fields = {
            "items": "fields",  # The data returned by the server is fields, which are reserved for pydantic.
        }


class PostDatasheetMetaResponse(ResponseBase):
    data: PostDatasheetMetaResponseData


class PostEmbedLinkResponseData(BaseModel):
    payload: Optional[EmbedLinkPayload]
    theme: Optional[EmbedLinkThemeEnum]
    linkId: str
    url: str


class PostEmbedLinkResponse(ResponseBase):
    data: PostEmbedLinkResponseData


GetEmbedLinkResponseData = PostEmbedLinkResponseData


class GetEmbedLinkResponse(ResponseBase):
    data: List[GetEmbedLinkResponseData]


class DeleteEmbedLinkResponse(ResponseBase):
    pass



import unittest
import warnings

from vika import Vika
from . import TOKEN, DOMAIN, SPACE_ID, DATASHEET_ID


class TestGetFields(unittest.TestCase):
    def setUp(self):
        warnings.simplefilter("ignore", ResourceWarning)
        apitable = Vika(TOKEN)
        apitable.set_api_base(DOMAIN)
        self.dst = apitable.space(SPACE_ID).datasheet(DATASHEET_ID)

    def test_fields_all(self):
        fields = self.dst.fields.all()
        self.assertIsInstance(fields, list)
        primary_field = fields[0]
        self.assertTrue(primary_field.isPrimary)
        self.assertEqual(primary_field.name, "title")
        self.assertEqual(primary_field.type, "SingleText")

    def test_field_get(self):
        primary_field = self.dst.fields.get("title")
        self.assertTrue(primary_field.isPrimary)
        self.assertEqual(primary_field.name, "title")
        self.assertEqual(primary_field.type, "SingleText")


if __name__ == "__main__":
    unittest.main()



import time
import unittest
import warnings
from vika import Vika
from . import DOMAIN, TOKEN, SPACE_ID, DATASHEET_ID


class TestEmbedLinkCRD(unittest.TestCase):

    def setUp(self):
        warnings.simplefilter("ignore", ResourceWarning)
        apitable = Vika(TOKEN)
        apitable.set_api_base(DOMAIN)
        self.spc = apitable.space(SPACE_ID)

    def test_embed_link_crd(self):
        datasheet = self.spc.datasheet(DATASHEET_ID)
        newEmbedLink = datasheet.create_embed_link()
        self.assertIsNotNone(newEmbedLink.linkId)
        time.sleep(0.5)
        embedLinks = datasheet.get_embed_links()
        self.assertEqual(len(embedLinks), 1)
        time.sleep(0.5)
        isDeleted = datasheet.delete_embed_link(newEmbedLink.linkId)
        self.assertTrue(isDeleted)
        time.sleep(0.5)
        embedLinks = datasheet.get_embed_links()
        self.assertEqual(len(embedLinks), 0)

    def tearDown(self):
        """Create an embed link test class object destroy hook:"""


if __name__ == "__main__":
    unittest.main()



from enum import Enum
from typing import Optional

from pydantic import BaseModel


class ViewTypeEnum(str, Enum):
    Grid = 'Grid'
    Gallery = 'Gallery'
    Kanban = 'Kanban'
    Gantt = 'Gantt'
    Calendar = 'Calendar'
    Architecture = 'Architecture'


class MetaView(BaseModel):
    id: str
    name: str
    type: Optional[ViewTypeEnum]



from functools import lru_cache, wraps
from json import JSONDecodeError
from time import monotonic
from typing import Any, Dict, Generic, TypeVar
from urllib.parse import urlparse

from vika.exceptions import ResponseBodyParserError, ServerError

T = TypeVar('T')

FieldKeyMap = Dict[str, str]


def chunks(_list, chunk_size):
    for i in range(0, len(_list), chunk_size):
        yield _list[i: i + chunk_size]


def trans_key(field_key_map: FieldKeyMap, key: str):
    """
    When there is a field mapping, convert the mapped key to the actual key
    """
    if key in ["_id", "recordId"]:
        return key
    if field_key_map:
        _key = field_key_map.get(key, key)
        return _key
    return key


def trans_data(field_key_map: FieldKeyMap, data: Dict[str, Any]):
    """
    After configuring the field mapping relationship, convert the incoming record data into data understood by the API
    """
    if field_key_map:
        _data = {}
        for k, v in data.items():
            _k = field_key_map.get(k, k)
            _data[_k] = v
        return _data
    return data


def query_parse(field_key_map: FieldKeyMap, **kwargs):
    """
    Convert query conditions to filterByFormula
    records.filter(title="hello", subtitle="world") => 'AND({subtitle}="world",{title}="hello")'
    1. The query through the filter and get parameters can only be converted into and conditions。
    """
    query_list = []
    for k, v in kwargs.items():
        # Handling null
        if v is None:
            v = 'BLANK()'
        # Handling string
        elif isinstance(v, str):
            v = f'"{v}"'
        elif isinstance(v, bool):
            v = 'TRUE()' if v else 'FALSE()'
        # Handling array type values, multiple select, members?
        elif isinstance(v, list):
            v = f'"{", ".join(v)}"'
        query_list.append(f'{{{trans_key(field_key_map, k)}}}={v}')
    if len(query_list) == 1:
        return query_list[0]
    else:
        qs = ",".join(query_list)
        return f"AND({qs})"


def handle_response(resp, resp_class: Generic[T]) -> T:
    if resp.status_code >= 500:
        raise ServerError(f"API Server Error: {resp.status_code}")
    try:
        r = resp.json()
        if r["success"]:
            try:
                return resp_class(**r)
            except:
                raise ResponseBodyParserError(f"Response Body Parser Error: {resp.text}")
        raise Exception(r['message'])
    except JSONDecodeError:
        raise Exception(f"JSON Parser Error: {resp.text}")


def check_sort_params(sort):
    if not isinstance(sort, list):
        return False
    return all([('field' in i and 'order' in i) for i in sort])


def get_dst_id(dst_id_or_url: str):
    if dst_id_or_url.startswith("dst"):
        return dst_id_or_url
    elif dst_id_or_url.startswith("http"):
        url = urlparse(dst_id_or_url)
        url_path_list = url.path.split("/")
        dst_id = url_path_list[-2]
        return dst_id
    else:
        raise Exception('Bad Datasheet Id or URL')


def timed_lru_cache(
    _func=None, *, seconds: int = 600, maxsize: int = 128, typed: bool = False
):
    """Extension of functools lru_cache with a timeout
    https://gist.github.com/Morreski/c1d08a3afa4040815eafd3891e16b945#gistcomment-3521580
    Parameters:
    seconds (int): Timeout in seconds to clear the WHOLE cache, default = 10 minutes
    maxsize (int): Maximum Size of the Cache
    typed (bool): Same value of different type will be a different entry

    """

    def wrapper_cache(f):
        f = lru_cache(maxsize=maxsize, typed=typed)(f)
        f.delta = seconds
        f.expiration = monotonic() + f.delta

        @wraps(f)
        def wrapped_f(*args, **kwargs):
            if monotonic() >= f.expiration:
                f.cache_clear()
                f.expiration = monotonic() + f.delta
            return f(*args, **kwargs)

        wrapped_f.cache_info = f.cache_info
        wrapped_f.cache_clear = f.cache_clear
        return wrapped_f

    # To allow decorator to be used without arguments
    if _func is None:
        return wrapper_cache
    else:
        return wrapper_cache(_func)



from urllib.parse import urljoin

from vika.types.response import (
    GETNodeListResponse,
    GETNodeDetailResponse
)
from vika.utils import handle_response


class NodeManager:
    def __init__(self, apitable: 'Apitable', **kwargs):
        self.apitable = apitable
        self.space_id = kwargs.get('space_id', None)

    def _get_nodes(self, space_id: str) -> GETNodeListResponse:
        """
        Get field meta
        """
        node_list_resp = self.apitable.request.get(
            urljoin(self.apitable.api_base, f"/fusion/v1/spaces/{space_id}/nodes")
        )
        return handle_response(node_list_resp, GETNodeListResponse)

    def _get_node_detail(self, node_id: str) -> GETNodeDetailResponse:
        node_detail_resp = self.apitable.request.get(
            urljoin(self.apitable.api_base, f"/fusion/v1/nodes/{node_id}")
        )
        return handle_response(node_detail_resp, GETNodeDetailResponse)

    def all(self, **kwargs):
        """
        apitable.nodes.all(spaceId='spcxxxxxx') Get the list of files in the root of the specified space
        """
        space_id = kwargs.get('spaceId', self.space_id)
        nodes_resp = self._get_nodes(space_id)
        return nodes_resp.data.nodes

    def get(self, node_id: str):
        return self._get_node_detail(node_id).data



from vika.types.response import PostDatasheetMetaResponseData


class DatasheetManager:
    """

    """
    def __init__(self, spc: 'Space'):
        self.spc = spc

    def create(self, data) -> PostDatasheetMetaResponseData:
        """ Create Field  https://developers.apitable.com/api/reference/#operation/create-datasheets

            :param dict data: 
            :return: datasheet id, Creation timestamp, field id and field name
            :raises ServerError
            :raises ResponseBodyParserError: Failed to parse response body
            :raises Exception: Other error

            :example:
            >>> apitable = Apitable('YOUR_API_TOKEN')
            >>> req_data = {'name': 'table_name'}
             >>> dst_meta = apitable.space('YOUR_SPACE_ID').datasheets.create(req_data)
        """
        resp = self.spc.create_datasheet(data)
        return resp.data



from enum import Enum
from pydantic import BaseModel
from typing import Optional


class EmbedLinkThemeEnum(str, Enum):
    Light = 'light'
    DARK = 'dark'


class EmbedLinkPermissionType(str, Enum):
    READ_ONLY = 'readOnly'
    PUBLIC_EDIT = 'publicEdit'
    PRIVATE_EDIT = 'privateEdit'


class EmbedLinkPayloadSideBar(BaseModel):
    collapsed: Optional[bool]


class EmbedLinkPayloadViewToolBar(BaseModel):
    basicTools: Optional[bool]
    shareBtn: Optional[bool]
    widgetBtn: Optional[bool]
    apiBtn: Optional[bool]
    formBtn: Optional[bool]
    historyBtn: Optional[bool]
    robotBtn: Optional[bool]


class EmbedLinkPayloadViewControl(BaseModel):
    viewId: Optional[str]
    tabBar: Optional[bool]
    toolBar: Optional[EmbedLinkPayloadViewToolBar]
    collapsed: Optional[bool]


class EmbedLinkPayload(BaseModel):
    primarySideBar: Optional[EmbedLinkPayloadSideBar]
    viewControl: Optional[EmbedLinkPayloadViewControl]
    collapsed: Optional[bool]
    permissionType: Optional[EmbedLinkPermissionType]


class EmbedLinkCreateRo(BaseModel):
    payload: Optional[EmbedLinkPayload]
    theme: Optional[EmbedLinkThemeEnum]




import unittest
import warnings
from vika import Vika
from . import TOKEN, DOMAIN, SPACE_ID, DATASHEET_ID


class TestDeleteFields(unittest.TestCase):
    """Apitable Python SDK Field delete test class
    - Field delete SDK test
    """

    def setUp(self):
        """Create a field test class object initial hook:
        - Warnings related to resource usage are ignored
        - Initialize the Apitable object
        - Get action sheet
        - add fields
        """
        warnings.simplefilter("ignore", ResourceWarning)
        apitable = Vika(TOKEN)
        apitable.set_api_base(DOMAIN)
        self.dst = apitable.space(SPACE_ID).datasheet(DATASHEET_ID)
        req_data = {
            "type": "SingleText",
            "name": "test_field_delete",
            "property": {"defaultValue": "hello apitable"},
        }
        self.field = self.dst.fields.create(req_data)

    def test_field_delete(self):
        """Field delete SDK test:"""
        is_true = self.dst.fields.delete(self.field.id)
        self.assertTrue(is_true)


if __name__ == "__main__":
    unittest.main()



import requests

from .const import API_BASE
from .datasheet import Datasheet
from .node import NodeManager
from .space import Space, SpaceManager
from .utils import get_dst_id


class Vika:
    def __init__(self, token: str, **kwargs):
        """
        @param token: API Token, User's developer token
        @param kwargs:
            - api_base: api url，default is "https://vika.cn"
        """
        self.request = requests.session()
        self._auth(token)
        self._api_base = kwargs.get('api_base', API_BASE)

    def _auth(self, token):
        self.request.headers.update({"Authorization": f"Bearer {token}"})

    @property
    def api_base(self):
        """
        Current API base address, distinguishing between environments
        """
        return self._api_base

    def set_api_base(self, api_base):
        """
        set api_base
        @param api_base:
        @return:
        """
        self._api_base = api_base

    def space(self, space_id: str):
        """
        Specify space
        @param space_id
        @return: Space
        """
        return Space(self, space_id)

    @property
    def spaces(self):
        """
        Space Resource Management
        @return: SpaceManager
        """
        return SpaceManager(self)

    @property
    def nodes(self):
        """
        Node Resource Management
        @return: NodeManager
        """
        return NodeManager(self)

    def datasheet(self, dst_id_or_url, **kwargs):
        """
        Specifying a Datasheet in Apitable
        @param dst_id_or_url: Datasheet ID or URL
        @param kwargs:
            - field_key: 'id' or 'name'
            - field_key_map: Field mapping. More info: https://github.com/apitable/apitable-sdks/tree/develop/apitable.py#Field-mapping
        @return:
        """
        dst_id = get_dst_id(dst_id_or_url)
        return Datasheet(self, dst_id, records=[], **kwargs)



API_BASE = "https://vika.cn"

API_ENDPOINT_DATASHEET = "/fusion/v1/datasheets"

MAX_WRITE_RECORDS_PRE_REQ = 10

MAX_GET_RECORDS_PRE_REQ = 1000

API_GET_DATASHEET_QS_SET = {
    "viewId",
    "pageNum",
    "pageSize",
    "sort",
    "recordIds",
    "fields",
    "filterByFormula",
    "maxRecords",
    "cellFormat",
    "fieldKey",
}

DEFAULT_PAGE_SIZE = 1000

QPS = 5



from vika.node import NodeManager
from vika.utils import get_dst_id, handle_response
from vika.datasheet import Datasheet, DatasheetManager
from vika.types.response import PostDatasheetMetaResponse
from urllib.parse import urljoin


class Space:
    def __init__(self, apitable: 'Apitable', space_id: str):
        self.apitable = apitable
        self.id = space_id

    @property
    def nodes(self):
        return NodeManager(self.apitable, space_id=self.id)

    @property
    def datasheets(self):
        return DatasheetManager(self)

    def datasheet(self, dst_id_or_url, **kwargs):
        """
        @param dst_id_or_url: Datasheet ID or URL
        @param kwargs:
            - field_key: 'id' or 'name' 
            - field_key_map: Field Mapping Dictionary. More info: https://github.com/apitable/apitable-sdks/tree/develop/apitable.py#Field-mapping
        @return:
        """
        dst_id = get_dst_id(dst_id_or_url)
        return Datasheet(self.apitable, dst_id, spc_id=self.id, **kwargs)

    def create_datasheet(self, data) -> PostDatasheetMetaResponse:
        """ 
            :param dic data: API request body, structure: {'name': 'table_name'}
            :return: Create form response data
            :raises ServerError:
            :raises ResponseBodyParserError: Failed to parse response body
            :raises Exception: Other error
        """
        if self.id is None:
            raise Exception('maybe: apitable.datasheet("dst_id") => apitable.space("spc_id").datasheet("dst_id")')
        api_endpoint = urljoin(self.apitable.api_base,
                               f"/fusion/v1/spaces/{self.id}/datasheets")
        resp = self.apitable.request.post(api_endpoint, json=data)
        return handle_response(resp, PostDatasheetMetaResponse)



import unittest
import time
import warnings
from vika import Vika
from . import TOKEN, DOMAIN, SPACE_ID, DATASHEET_ID


class TestCreateRecords(unittest.TestCase):
    def setUp(self):
        warnings.simplefilter("ignore", ResourceWarning)
        apitable = Vika(TOKEN)
        apitable.set_api_base(DOMAIN)
        self.dst = apitable.space(SPACE_ID).datasheet(DATASHEET_ID)

    def test_record_create(self):
        time.sleep(1)
        record = self.dst.records.create({"title": "Advanced Mathematics"})
        time.sleep(1)
        self.assertIsNotNone(record._id)
        records = self.dst.records.bulk_create(
            [{"title": "Discrete Mathematics"}, {"title": "Linear Algebra"}]
        )
        self.created_records = records + [record]
        for rec in records:
            self.assertIsNotNone(rec._id)

    def tearDown(self):
        self.dst.delete_records(self.created_records)


if __name__ == "__main__":
    unittest.main()



from enum import Enum

from typing import List, Optional
from pydantic import BaseModel


class NodeTypeEnum(str, Enum):
    Datasheet = 'Datasheet'
    Folder = 'Folder'
    Form = 'Form'
    Dashboard = 'Dashboard'
    Mirror = 'Mirror'


class NodeListItem(BaseModel):
    id: str
    name: str
    type: NodeTypeEnum
    icon: str
    isFav: bool


class NodeDetail(NodeListItem):
    type = NodeTypeEnum.Folder
    children: Optional[List[NodeListItem]]



from typing import List

from vika.const import MAX_GET_RECORDS_PRE_REQ, MAX_WRITE_RECORDS_PRE_REQ
from vika.datasheet.query_set import QuerySet
from vika.datasheet.record import Record
from vika.exceptions import RecordDoesNotExist
from vika.types import GETRecordResponse
from vika.utils import query_parse, trans_data


class RecordManager:

    def __init__(self, dst: 'Datasheet'):
        self._dst = dst
        self._fetched_with = None
        self._fetched_by = None

    def bulk_create(self, data) -> List[Record]:
        """
        Create records in batches, only 10 records can be created per request dst.records.bulk_create([{"title": "hello apitable"}])
        @param data: list of record objects [{ "fieldKey": fieldValue }, { "fieldKey": fieldValue2 }]
        @return: List[Record]
        """
        if len(data) > MAX_WRITE_RECORDS_PRE_REQ:
            raise Exception(f'The number of records created by a single request cannot be greater than {MAX_WRITE_RECORDS_PRE_REQ}')
        resp = self._dst.create_records(data)
        return [Record(self._dst, record) for record in resp.data.records]

    def bulk_update(self, data) -> List[Record]:
        """
        Batch update records, same as raw API request
        @param data: record object list [{ "recordId": "recxxxx", "fields": {"fieldKey": fieldValue} }]
        @return Boolean
        """
        if len(data) > MAX_WRITE_RECORDS_PRE_REQ:
            raise Exception(f'The number of records to be updated in a single request cannot be greater than {MAX_WRITE_RECORDS_PRE_REQ}')
        for record in data:
            record['fields'] = trans_data(self._dst.field_key_map,
                                          record['fields'])
        updated_records = self._dst.update_records(data)
        return [Record(self._dst, record) for record in updated_records]

    def create(self, data) -> Record:
        """
        Create a record dst.records.create({"title": "hello apitable"})
        @param data: dict {"fieldKey": fieldValue}
        @return: Record
        """
        resp = self._dst.create_records(data)
        if resp.success:
            records = resp.data.records
            return Record(self._dst, records[0])
        raise Exception(resp.message)

    def all(self, **kwargs):
        """
        Filter records by query conditions, see get records query params https://developers.apitable.com/api-reference#operation/get-records
        @param kwargs:
            - viewId: 'viewId1',View ID. Defaults to the first view in the Apitable. The request will return the filtered/sorted results in the view, and you can use the fields parameter to filter unwanted field data
            - sort: [{ 'Field name or ID': 'asc' }] sorts the records of the specified dimensional grid. An array of multiple "sort objects". Support order: 'asc' and reverse order: 'desc'. Note: The sorting conditions specified by this parameter will override the sorting conditions in the view.
            - recordIds: ['recordId1', 'recordId2'], array of recordIds. If this parameter is included, the records array specified in the parameter is returned. The returned values are sorted in the order passed in the array. Filtering and sorting are ignored at this time. No paging, at most 1000 items per query
            - fields: ['title', 'detail', 'Citations'], specify the field to be returned (the default is the field name, or it can be specified as the field Id through fieldKey). If this parameter is included, the returned record collection will be filtered, and only the specified fields will be returned.
            - filterByFormula: '{Citations} >  0', use formulas as filter conditions to return matching records, visit https://help.apitable.com/docs/guide/tutorial-getting-started-with-formulas to learn how to use formulas
            - maxRecords: 5000, limits the total number of records returned. If this value is less than the actual total number of records in the table, the total number of records returned is limited to this value.
            - cellFormat: 'json', cell value type, the default is 'json', when specified as 'string', all values will be automatically converted to string format.
            - fieldKey: 'name', specifies the field's query and returned key. The column name 'name' is used by default. When specified as 'id', fieldId will be used as the query and return method (using id can avoid the problem of code failure caused by modification of column names)
        @return:
        """
        _fieldKey = kwargs.get("fieldKey")
        if _fieldKey and _fieldKey != self._dst.field_key:
            # TODO: logger warning
            print(
                f'It seems that you set field_key when init datasheet, all(filedKey="{_fieldKey}") wont work'
            )
        kwargs.update(fieldKey=self._dst.field_key)
        if 'pageSize' in kwargs or 'pageNum' in kwargs:
            resp: GETRecordResponse = self._dst.get_records(**kwargs)
            if resp.success:
                records = resp.data.records
            else:
                print(f"[{self._dst.id}] fetch data fail\n {resp.message}")
                records = []
        else:
            records = self._dst.get_records_all(**kwargs)
        return QuerySet(self._dst, records)

    def get(self, *args, **kwargs):
        """
        Query a single record that meets the conditions, suitable for querying using a uniquely identified field
        1. Specify record id to query records
        dst_books.records.get("recxxxxxx")
        2. Query records by condition
        dst_books.records.get(ISBN="9787506341271")
        @param args:
        @param kwargs:
        @return:
        """
        if args:
            record_id = args[0]
            kwargs = {"recordIds": [record_id]}
        elif kwargs:
            query_formula = query_parse(self._dst.field_key_map, **kwargs)
            kwargs = {"filterByFormula": query_formula}
        resp: GETRecordResponse = self._dst.get_records(**kwargs)
        if resp.data.records:
            return Record(self._dst, resp.data.records[0])
        raise RecordDoesNotExist()

    def get_or_create(self, defaults=None, **kwargs):
        """
        Find a single record by the given kwargs criteria, and create a record if no matching record exists.
        eg: dst.records.get_or_create(defaults={"done":False}, task="Find apitable") If a record with "task" as "Find apitable" is not found, create one with "done" of "False"
        @param defaults Dictionary (field key, field value), when the record does not exist, create the default value of the record.
        @param kwargs Query conditions
        @return (record,created) Returns the record and whether it was created
        """
        created = False
        record = None
        data = {}
        try:
            record = self.get(**kwargs)
        except RecordDoesNotExist:
            created = True
            if kwargs:
                data.update(kwargs)
            if defaults:
                data.update(defaults)
            record = self.create(data)
        return (record, created)

    def update_or_create(self, defaults=None, **kwargs):
        """
        Find a single record by the given kwargs criteria, and create a record if no matching record exists. If found, update the record with the value in defaults.
        @param defaults Dictionary (field key, field value), when the record does not exist, create the default value of the record.
        @param kwargs Query conditions
        @return (record,created) Returns the record and whether it was created
        """
        created = False
        record = None
        data = {}
        try:
            record = self.get(**kwargs).update(defaults)
        except RecordDoesNotExist:
            created = True
            if kwargs:
                data.update(kwargs)
            if defaults:
                data.update(defaults)
            record = self.create(data)
        return (record, created)

    def filter(self, **kwargs):
        """
        Through the query conditions, query the recordsets that meet the conditions
        songs = dst_songs.records.filter(artist="faye wong")
        for song in songs:
            print(song.title)
        @param kwargs:
        @return: QuerySet
        """
        # When calling directly through filter, convert the filter query parameter into filterByFormula, and use the server to calculate the result
        records = self._query_records(**kwargs)
        return QuerySet(self._dst, records)

    def _query_records(self, **kwargs):
        # Convert the query conditions into filterByFormula, and use the server to calculate the query record set
        query_formula = query_parse(self._dst.field_key_map, **kwargs)
        kwargs = {
            "filterByFormula": query_formula,
            "pageSize": MAX_GET_RECORDS_PRE_REQ
        }
        return self._dst.get_records_all(**kwargs)



from typing import Dict, Any
from pydantic import BaseModel


# Avoid the same name as the Record class
class RawRecord(BaseModel):
    """
    The record primitive type returned by the REST API
    """

    id: str
    data: Dict[str, Any]
    # createdAt: int
    # updatedAt: int

    class Config:
        # https://github.com/samuelcolvin/pydantic/issues/1250
        fields = {
            "data": "fields",  # The data returned by the server is fields, which are reserved for pydantic.
            "id": "recordId",  # The data returned by the server is recordId, which is mapped to id here
        }



import unittest
import warnings

from vika import Vika
from . import TOKEN, DOMAIN, SPACE_ID, DATASHEET_ID


class TestGetViews(unittest.TestCase):
    def setUp(self):
        warnings.simplefilter("ignore", ResourceWarning)
        apitable = Vika(TOKEN)
        apitable.set_api_base(DOMAIN)
        self.dst = apitable.space(SPACE_ID).datasheet(DATASHEET_ID)

    def test_views_all(self):
        views = self.dst.views.all()
        self.assertIsInstance(views, list)
        first_view = views[0]
        self.assertEqual(first_view.name, "Grid View")
        self.assertEqual(first_view.type, "Grid")


if __name__ == "__main__":
    unittest.main()



from urllib.parse import urljoin

from vika.types.response import (
    GETSpaceListResponse
)
from vika.utils import handle_response


class SpaceManager:
    def __init__(self, apitable: 'Apitable', **kwargs):
        self.apitable = apitable

    @property
    def _api_endpoint(self):
        return urljoin(self.apitable.api_base, f"/fusion/v1/spaces")

    def _get_spaces(self) -> GETSpaceListResponse:
        """
        Get field meta
        """
        resp = self.apitable.request.get(self._api_endpoint)
        return handle_response(resp, GETSpaceListResponse)

    def all(self):
        """
        Get the current user's space list
        @return: List[SpaceListItem]
        """
        spaces_resp = self._get_spaces()
        return spaces_resp.data.spaces

    def get(self, space_id: str):
        """
        Get the specified space information
        @param space_id
        @return:
        """
        pass



class DatasheetDoesNotExist(Exception):
    """
    Datasheet does not exist
    """
    pass


class RecordDoesNotExist(Exception):
    """
    Record does not exist
    """
    pass


class FieldDoesNotExist(Exception):
    """
    Field does not exist
    """

    pass


class RecordWasDeleted(Exception):
    """
    After the record is deleted, an error will be reported when accessing the properties of this record again.
    """

    pass


class ErrorFieldKey(Exception):
    """
    Wrong fieldKey
    """

    pass


class ErrorSortParams(Exception):
    """
    Wrong sorting rules
    """
    pass


class UploadFileError(Exception):
    """
    Failed to upload attachment
    """
    pass


class JSONDecodeError(Exception):
    """
    JSON parsing error
    """
    pass


class ResponseBodyParserError(Exception):
    """
    Failed to analyze the response body
    """
    pass


class ServerError(Exception):
    """
    Server error
    """
    pass



import io
import json
import mimetypes
from typing import List
from urllib.parse import urljoin

import requests

from vika.const import API_GET_DATASHEET_QS_SET, DEFAULT_PAGE_SIZE
from vika.exceptions import ErrorSortParams
from vika.datasheet.field_manager import FieldManager
from vika.datasheet.record import Record
from vika.datasheet.record_manager import RecordManager
from vika.types import EmbedLinkCreateRo
from vika.types.response import (
    GETMetaFieldResponse,
    PostMetaFieldResponse,
    PatchRecordResponse,
    PostRecordResponse,
    DeleteRecordResponse,
    UploadFileResponse,
    RawRecord,
    GETMetaViewResponse,
    GETRecordResponse,
    DeleteFieldResponse,
    PostEmbedLinkResponse,
    GetEmbedLinkResponse,
    PostEmbedLinkResponseData,
    GetEmbedLinkResponseData,
    DeleteEmbedLinkResponse,
)
from vika.utils import FieldKeyMap, handle_response, check_sort_params, trans_data, timed_lru_cache
from vika.datasheet.view_manager import ViewManager


class Datasheet:

    def __init__(self, apitable: 'Apitable', dst_id: str, spc_id=None, **kwargs):
        self.apitable = apitable
        self.id = dst_id
        self.spc_id = spc_id
        field_key = kwargs.get("field_key", "name")
        if field_key not in ["name", "id"]:
            raise Exception("Error field_key, plz use「name」 or 「id」")
        self.field_key = field_key
        field_key_map = kwargs.get("field_key_map", None)
        self.field_key_map: FieldKeyMap = field_key_map

    def refresh(self):
        """
        TODO refetch datasheet
        """
        pass

    @property
    def _record_api_endpoint(self):
        return urljoin(self.apitable.api_base,
                       f"/fusion/v1/datasheets/{self.id}/records")

    @property
    @timed_lru_cache(seconds=300)
    def fields(self):
        return FieldManager(self)

    @property
    def primary_field(self):
        return self.fields.all()[0]

    @property
    def views(self):
        return ViewManager(self)

    @property
    def records(self):
        return RecordManager(self)

    # The following is a request for datasheet management
    # Field
    def get_fields(self, **kwargs):
        """
        Get field meta
        @param kwargs:
            - viewId: 'viewId'
        @return:
        """
        api_endpoint = urljoin(self.apitable.api_base,
                               f"/fusion/v1/datasheets/{self.id}/fields")
        resp = self.apitable.request.get(api_endpoint, params=kwargs)
        return handle_response(resp, GETMetaFieldResponse)

    def create_field(self, data) -> PostMetaFieldResponse:
        """ 
            :param dic data: {'type': str, 'name': str, 'property': obj}
            :return: Field creation api returns result
            :raises ServerError
            :raises ResponseBodyParserError: Failed to parse response body
            :raises Exception: Other error, such as: field duplicate name
        """
        if self.spc_id is None:
            raise Exception('maybe: apitable.datasheet("dst_id") => apitable.space("spc_id").datasheet("dst_id")')
        api_endpoint = urljoin(self.apitable.api_base,
                               f"/fusion/v1/spaces/{self.spc_id}/datasheets/{self.id}/fields")
        resp = self.apitable.request.post(api_endpoint, json=data)
        return handle_response(resp, PostMetaFieldResponse)

    def delete_field(self, field_id: str) -> bool:
        """ 
            :param str field_id
            :return: Whether the field was deleted successfully
            :raises ServerError
            :raises ResponseBodyParserError: Failed to parse response body
            :raises Exception: Other error
        """
        if self.spc_id is None:
            raise Exception('maybe: apitable.datasheet("dst_id") => apitable.space("spc_id").datasheet("dst_id")')
        api_endpoint = urljoin(self.apitable.api_base,
                               f"/fusion/v1/spaces/{self.spc_id}/datasheets/{self.id}/fields/{field_id}")
        resp = self.apitable.request.delete(api_endpoint)
        r = handle_response(resp, DeleteFieldResponse)
        return r.success

    # View
    def get_views(self):
        """
        Get view meta
        """
        api_endpoint = urljoin(self.apitable.api_base,
                               f"/fusion/v1/datasheets/{self.id}/views")
        r = self.apitable.request.get(api_endpoint).json()
        r = GETMetaViewResponse(**r)
        if r.success:
            return r.data.views
        raise Exception(r.message)

    # Record
    def get_records(self, **kwargs):
        """
        Paginate to get data
        """
        params = {}
        for key in kwargs:
            if key in API_GET_DATASHEET_QS_SET:
                params_value = kwargs.get(key)
                if key == 'sort':
                    if check_sort_params(params_value):
                        params_value = [json.dumps(i) for i in params_value]
                    else:
                        raise ErrorSortParams('The format of the sort parameter is incorrect')
                params.update({key: params_value})
        resp = self.apitable.request.get(self._record_api_endpoint, params=params)
        return handle_response(resp, GETRecordResponse)

    def get_records_all(self, **kwargs) -> List[RawRecord]:
        """
        When pageSize and pageNum are not actively passed in, all records will be loaded.
        """
        page_size = kwargs.get("pageSize", DEFAULT_PAGE_SIZE)
        page_num = kwargs.get("pageNum", 1)
        page_params = {"pageSize": page_size, "pageNum": page_num}
        kwargs.update(page_params)
        records = []
        resp = self.get_records(**kwargs)
        if resp.success:
            records += resp.data.records
            current_total = page_size * page_num
            if current_total < resp.data.total:
                kwargs.update({"pageNum": page_num + 1})
                records += self.get_records_all(**kwargs)
        else:
            print(f"[{self.id}] get page:{page_num} fail\n {resp.message}")
        return records

    def create_records(self, data) -> PostRecordResponse:
        if type(data) is list:
            data = {
                "records": [{
                    "fields": trans_data(self.field_key_map, item)
                } for item in data],
                "fieldKey":
                    self.field_key,
            }
        else:
            data = {
                "records": [{
                    "fields": trans_data(self.field_key_map, data)
                }],
                "fieldKey": self.field_key,
            }
        resp = self.apitable.request.post(self._record_api_endpoint, json=data)
        return handle_response(resp, PostRecordResponse)

    def delete_records(self, rec_list) -> bool:
        api_endpoint = self._record_api_endpoint
        if type(rec_list) is list:
            ids = [rec._id if type(rec) is Record else rec for rec in rec_list]
        else:
            rec = rec_list
            ids = rec._id if type(rec) is Record else rec
        resp = self.apitable.request.delete(api_endpoint,
                                            params={"recordIds": ids})
        r = handle_response(resp, DeleteRecordResponse)
        return r.success

    def update_records(self, data) -> List[RawRecord]:
        if type(data) is list:
            data = {"records": data, "fieldKey": self.field_key}
        else:
            data = {"records": [data], "fieldKey": self.field_key}
        r = self.apitable.request.patch(self._record_api_endpoint,
                                        json=data).json()
        if r["success"]:
            r = PatchRecordResponse(**r)
            return r.data.records
        else:
            raise Exception(r["message"])

    def upload_attachment(self, file_url):
        """
        Upload attachments, support local or url.
        attachment = dst.upload_attachment("/path/to/your/file")
        record.attachment = [attachment]
        """
        return self.upload_file(file_url)

    def upload_file(self, file_url):
        """
        @deprecated
        pload attachments, support local or url.(Please use upload_attachment instead, this method may be deprecated)
        dst.upload_file("/path/to/your/file")
        """
        api_endpoint = urljoin(self.apitable.api_base,
                               f"/fusion/v1/datasheets/{self.id}/attachments")
        is_web_file = type(file_url) is str and file_url.startswith("http")

        if is_web_file:
            r = requests.get(file_url)
            file_mimetype = r.headers["content-type"]
            with io.BytesIO() as buf:
                buf.write(r.content)
                buf.seek(0)
                _file = ("image", io.BufferedReader(buf), file_mimetype)
                r = self.apitable.request.post(
                    api_endpoint,
                    files={"files": _file},
                    stream=False,
                )
        else:
            with open(file_url, "rb") as upload_file:
                r = self.apitable.request.post(
                    api_endpoint,
                    files={
                        "files": (
                            file_url,
                            upload_file,
                            mimetypes.guess_type(file_url)[0],
                        )
                    },
                )
        up_file_resp = handle_response(r, UploadFileResponse)
        if up_file_resp.success:
            return up_file_resp.data

    def create_embed_link(self, data: EmbedLinkCreateRo = None) -> PostEmbedLinkResponseData:
        api_endpoint = urljoin(self.apitable.api_base,
                               f"/fusion/v1/spaces/{self.spc_id}/nodes/{self.id}/embedlinks")
        resp = self.apitable.request.post(api_endpoint, json=data)
        return handle_response(resp, PostEmbedLinkResponse).data

    def get_embed_links(self) -> List[GetEmbedLinkResponseData]:
        api_endpoint = urljoin(self.apitable.api_base,
                               f"/fusion/v1/spaces/{self.spc_id}/nodes/{self.id}/embedlinks")
        resp = self.apitable.request.get(api_endpoint)
        return handle_response(resp, GetEmbedLinkResponse).data

    def delete_embed_link(self, link_id: str) -> bool:
        api_endpoint = urljoin(self.apitable.api_base,
                               f"/fusion/v1/spaces/{self.spc_id}/nodes/{self.id}/embedlinks/{link_id}")
        resp = self.apitable.request.delete(api_endpoint)
        r = handle_response(resp, DeleteEmbedLinkResponse)
        return r.success



import time
import unittest
import warnings

from vika import Vika

from . import TOKEN, DOMAIN, SPACE_ID, DATASHEET_ID


class TestDeleteRecords(unittest.TestCase):
    def setUp(self):
        warnings.simplefilter("ignore", ResourceWarning)
        apitable = Vika(TOKEN)
        apitable.set_api_base(DOMAIN)
        self.dst = apitable.space(SPACE_ID).datasheet(DATASHEET_ID)

    def test_record_delete(self):
        self.dst.records.create({"title": "apitable"})
        time.sleep(1)
        record = self.dst.records.get(title="apitable")
        time.sleep(1 / 5)
        r = record.delete()
        self.assertTrue(r)

        time.sleep(1 / 5)
        record_id = "recnotexist"
        with self.assertRaises(Exception) as e:
            self.dst.delete_records([record_id])
        self.assertIsNotNone(e)

    def tearDown(self):
        time.sleep(1)
        self.dst.records.create({"title": "apitable"})


if __name__ == "__main__":
    unittest.main()



from enum import Enum
from typing import List, Optional, Union, Any

from pydantic import BaseModel
from pydantic.typing import ForwardRef


class MemberEnum(str, Enum):
    Member = 'Member'
    Team = 'Team'


class RollUpFunctionEnum(str, Enum):
    VALUES = 'VALUES'
    AVERAGE = 'AVERAGE'
    COUNT = 'COUNT'
    COUNTA = 'COUNTA'
    COUNTALL = 'COUNTALL'
    SUM = 'SUM'
    MIN = 'MIN'
    MAX = 'MAX'
    AND = 'AND'
    OR = 'OR'
    XOR = 'XOR'
    CONCATENATE = 'CONCATENATE'
    ARRAYJOIN = 'ARRAYJOIN'
    ARRAYUNIQUE = 'ARRAYUNIQUE'
    ARRAYCOMPACT = 'ARRAYCOMPACT'


class ComputeValueTypeEnum(str, Enum):
    String = 'String'
    Boolean = 'Boolean'
    Number = 'Number'
    Array = 'Array'
    DateTime = 'DateTime'


# field property
class SingleTextFieldProperty(BaseModel):
    defaultValue: Optional[str]


class NumberFieldProperty(BaseModel):
    defaultValue: Optional[str]
    precision: int


class CurrencyFieldProperty(NumberFieldProperty):
    symbol: str


class PercentFieldProperty(NumberFieldProperty):
    pass


class SelectOptionColor(BaseModel):
    name: str
    value: str


class SelectOption(BaseModel):
    id: str
    name: str
    color: SelectOptionColor


class SingleSelectFieldProperty(BaseModel):
    options: List[SelectOption]


class MultiSelectFieldProperty(SingleSelectFieldProperty):
    pass


class MemberOption(BaseModel):
    id: str
    name: str
    type: MemberEnum
    avatar: Optional[str]


class MemberFieldProperty(BaseModel):
    options: List[MemberOption]


class UserOption(BaseModel):
    id: str
    name: str
    avatar: str


class CreatedByFieldProperty(BaseModel):
    options: List[UserOption]


class LastModifiedByFieldProperty(CreatedByFieldProperty):
    pass


class CheckboxFieldProperty(BaseModel):
    icon: str


class RatingFieldProperty(BaseModel):
    icon: str
    max: int


class DateTimeFieldProperty(BaseModel):
    format: str
    autoFill: bool = False
    includeTime: bool = False


class CreatedTimeFieldProperty(DateTimeFieldProperty):
    pass


class LastModifiedTimeFieldProperty(DateTimeFieldProperty):
    pass


class MagicLinkFieldProperty(BaseModel):
    foreignDatasheetId: str
    brotherFieldId: Optional[str] = ""  # No brother field when it links itself


class FieldPropertyWithDstId(BaseModel):
    datasheetId: str
    field: ForwardRef("MetaField")


class MagicLookupFieldProperty(BaseModel):
    relatedLinkFieldId: str
    targetFieldId: str
    hasError: Optional[bool]
    entityField: Optional[FieldPropertyWithDstId]
    rollupFunction: RollUpFunctionEnum
    valueType: Optional[ComputeValueTypeEnum] = ComputeValueTypeEnum.String


class FormulaFieldProperty(BaseModel):
    expression: Optional[str]
    valueType: Optional[ComputeValueTypeEnum] = ComputeValueTypeEnum.String
    hasError: Optional[bool]


FieldProperty = Union[
    NumberFieldProperty,
    SingleTextFieldProperty,
    CurrencyFieldProperty,
    PercentFieldProperty,
    SingleSelectFieldProperty,
    MultiSelectFieldProperty,
    MemberFieldProperty,
    CreatedByFieldProperty,
    LastModifiedByFieldProperty,
    CheckboxFieldProperty,
    RatingFieldProperty,
    DateTimeFieldProperty,
    CreatedTimeFieldProperty,
    LastModifiedTimeFieldProperty,
    MagicLinkFieldProperty,
    MagicLookupFieldProperty,
    FormulaFieldProperty,
]


# field item
class MetaField(BaseModel):
    id: str
    name: str
    type: str
    isPrimary: Optional[bool] = False
    desc: Optional[str]
    editable: Optional[bool]

    # Union Types is not parsed correctly, here is the manual parsing of Any.
    # fuck https://github.com/samuelcolvin/pydantic/issues/2941
    property: Any = None

    def __init__(self, property=None, **data) -> None:
        property_model = self.get_property_by_type(data['type'])
        _property = property_model(**property) if property_model else property
        super().__init__(property=_property, **data)

    @staticmethod
    def get_property_by_type(type):
        type_property_map = {
            "SingleText": SingleTextFieldProperty,
            "Text": None,
            "SingleSelect": SingleSelectFieldProperty,
            "MultiSelect": MultiSelectFieldProperty,
            "Number": NumberFieldProperty,
            "Currency": CurrencyFieldProperty,
            "Percent": PercentFieldProperty,
            "DateTime": DateTimeFieldProperty,
            "Attachment": None,
            "Member": MemberFieldProperty,
            "Checkbox": CheckboxFieldProperty,
            "Rating": RatingFieldProperty,
            "URL": None,
            "Phone": None,
            "Email": None,
            "MagicLink": MagicLinkFieldProperty,
            "MagicLookUp": MagicLookupFieldProperty,
            "Formula": FormulaFieldProperty,
            "AutoNumber": None,
            "CreatedTime": CreatedTimeFieldProperty,
            "LastModifiedTime": LastModifiedTimeFieldProperty,
            "CreatedBy": CreatedByFieldProperty,
            "LastModifiedBy": LastModifiedByFieldProperty,
        }
        return type_property_map.get(type, None)


FieldPropertyWithDstId.update_forward_refs()



from vika.exceptions import RecordWasDeleted, ErrorFieldKey
from vika.types import RawRecord
from vika.utils import trans_key, trans_data


class Record:

    def __init__(self, datasheet: 'Datasheet', record: RawRecord):
        self._datasheet = datasheet
        self._id = record.id
        self._record = record
        self._is_del = False

    def _get_field(self, field_key: str):
        return self._datasheet.fields.get(field_key)

    def _check_record_status(self):
        if self._is_del:
            raise RecordWasDeleted()
        return None

    def __str__(self):
        return f"(Record: {self._id})"

    __repr__ = __str__

    def __getattr__(self, key):
        transformed_key = trans_key(self._datasheet.field_key_map, key)
        if not transformed_key:
            raise Exception(f"record has no field:[{key}]")
        # The data can be obtained, indicating that it has returned
        if transformed_key in self._record.data:
            return self._record.data.get(transformed_key)
        # I can't get it in the data, but this field exists. Indicates that the field value is empty
        if key in self._datasheet.fields:
            return None
        # Field does not exist
        raise ErrorFieldKey(f"'{key}' does not exist")

    def delete(self) -> bool:
        """
        @return: bool Whether the deletion is successful, the deletion exception will also throw an error
        """
        self._check_record_status()
        return self._datasheet.delete_records([self._id])

    def __setattr__(self, _key, value):
        if _key.startswith("_"):
            super().__setattr__(_key, value)
        elif _key in self._datasheet.fields:
            key = trans_key(self._datasheet.field_key_map, _key)
            data = {"recordId": self._id, "fields": {key: value}}
            updated_records = self._datasheet.update_records(data)
            if updated_records:
                self._record.data[key] = value
        else:
            raise ErrorFieldKey(f"'{_key}' does not exist")

    def json(self):
        self._check_record_status()
        # FIXME: Fill in the empty value field, use the original data, or use the mapped field name as the key
        record_data = dict(self._record.data)
        return record_data

    def _make_update_body(self, data):
        _data = trans_data(self._datasheet.field_key_map, data)
        data = {"recordId": self._id, "fields": _data}
        return data

    def update(self, data=None, **kwargs):
        """
       Update multiple fields, support two writing methods. You can pass a single dict, or you can pass in **kwargs
        @param data: dict The key-value pair that needs to be updated
        @param kwargs: The key-value pair that needs to be updated
        """
        self._check_record_status()
        update_data = {}
        if data:
            update_data.update(data)
        if kwargs:
            update_data.update(kwargs)
        # Update multiple fields of a single record, returning only one record
        update_data = self._make_update_body(update_data)
        self._record = self._datasheet.update_records(update_data)[0]
        return self



from typing import Optional, List, Dict, Any
from vika.types.field import MetaField
from vika.utils import trans_key


class FieldManager:

    def __init__(self, dst: 'Datasheet'):
        self.dst = dst
        self._is_fetched = False

        # All acquired fields information is cached once acquired
        self._fields = []
        # Supplementary data
        self._meta_field_id_map = {}
        self._meta_field_name_map = {}

    def refresh(self):
        """
        Refresh the field meta cache, the default cache is 5 minutes, and can be refreshed manually.
        """
        self._is_fetched = False

    def _check_meta(self, **kwargs):
        if not self._is_fetched:
            fields_resp = self.dst.get_fields(**kwargs)
            self._set_meta_fields(fields_resp.data.items)
            self._is_fetched = True

    def _set_meta_fields(self, fields: List[MetaField]):
        self._fields = fields
        for field in fields:
            self._meta_field_id_map[field.id] = field
            self._meta_field_name_map[field.name] = field

    def __contains__(self, item):
        """
        Is field_key in dst.fields
        @param item: field_key
        @return: bool
        """
        self._check_meta()
        if isinstance(item, MetaField):
            return item in self._fields
        field_key = trans_key(self.dst.field_key_map, item)
        return field_key in self._meta_field_id_map or field_key in self._meta_field_name_map

    def all(self, **kwargs) -> List[MetaField]:
        """
        Get all the field information of the current datasheet
        @param kwargs:
            - viewId: 'viewId' Get the field information of the specified view, the hidden fields under the view will not be returned.
        @return:
        """
        # Get the field information of the specified view, one-time without caching. Can it actually be cached?
        if kwargs:
            fields_resp_via_view_id = self.dst.get_fields(**kwargs)
            return fields_resp_via_view_id.data.items
        self._check_meta()
        return self._fields

    def get(self, field_key: str, **kwargs) -> Optional[MetaField]:
        """
        Obtained by field_key, the fields in the Apitable cannot have the same name, id and name are both unique identifiers
        dst.fields.get('fldxxxxxxxx')
        dst.fields.get('title')
        @param field_key
        @param kwargs:
        @return: MetaField
        """
        field_key = trans_key(self.dst.field_key_map, field_key)
        # If you haven't obtained meta, first request meta
        self._check_meta()
        # Get the meta and return the field directly
        if self.dst.field_key == "id":
            return self._meta_field_id_map.get(field_key, None)
        if self.dst.field_key == "name":
            return self._meta_field_name_map.get(field_key, None)
        # Field not found
        return None

    def create(self, data) -> Optional[MetaField]:
        """ Field creation https://developers.apitable.com/api/reference/#operation/create-fields

            :param dict data:  New field property
            :return: New field id and name
            :raises ServerError
            :raises ResponseBodyParserError: Failed to parse response body
            :raises Exception: Other error, such as: field duplicate name

            :example:
            >>> apitable = Apitable('YOUR_API_TOKEN')
            >>> req_data = {'type': 'SingleText', 'name': 'title', 'property': {'defaultValue': 'hello apitable'}}
            >>> fld_meta = apitable.space('YOUR_SPACE_ID').datasheet('YOUR_TABLE').fields.create(req_data)
        """
        resp = self.dst.create_field(data)
        self.refresh()
        return resp.data

    def delete(self, fid_id: str) -> bool:
        """ Field deletion https://developers.apitable.com/api/reference/#operation/delete-fields

            :param str fid_id: Field information to delete
            :return: Delete successfully returns true
            :raises ServerError
            :raises ResponseBodyParserError: Failed to parse response body
            :raises Exception: Other error, such as: field id does not exist

            :example:
            >>> apitable = Apitable('YOUR_API_TOKEN')
            >>> is_true = apitable.space('YOUR_SPACE_ID').datasheet('YOUR_TABLE').fields.delete('fid_id')
            >>> is_true
            True
        """
        is_success = self.dst.delete_field(fid_id)
        self.refresh()
        return is_success

    def __getitem__(self, index):
        self._check_meta()
        return self._fields[index]



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

with open(".version", "r") as fh:
    version_number = fh.read()

setuptools.setup(
    name="vika",
    version=version_number,
    author="vikadata",
    author_email="dev@vikadata.com",
    description="维格表官方 Python SDK",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/vikadata/vika.py",
    packages=[
        "vika", "vika.datasheet", "vika.types", "vika.node", "vika.space"
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
    install_requires=["requests", "pydantic", "environs"],
)



import unittest
import warnings
from vika import Vika
from . import DOMAIN, TOKEN, SPACE_ID, FOLDER_ID


class TestCreateFields(unittest.TestCase):
    """Apitable Python SDK create datasheet test class
    - Datasheet Creation SDK Test
    """

    def setUp(self):
        """Create a datasheet test class object initial hook:"""
        warnings.simplefilter("ignore", ResourceWarning)
        apitable = Vika(TOKEN)
        apitable.set_api_base(DOMAIN)
        self.spc = apitable.space(SPACE_ID)

    def test_datasheet_create(self):
        """Datasheet Creation SDK Test"""
        req_data = {"name": "python_test_datasheet_create", "folderId": FOLDER_ID}
        self.datasheet = self.spc.datasheets.create(req_data)
        self.assertIsNotNone(self.datasheet.id)

    def tearDown(self):
        """Create a Datasheet test class object destroy hook:"""


if __name__ == "__main__":
    unittest.main()



import unittest
import warnings

from vika import Vika

from . import TOKEN, DOMAIN, SPACE_ID, FOLDER_ID


class TestGetNodes(unittest.TestCase):
    def setUp(self):
        warnings.simplefilter("ignore", ResourceWarning)
        apitable = Vika(TOKEN)
        apitable.set_api_base(DOMAIN)
        self.apitable = apitable

    def test_space_root_nodes_all(self):
        space_root_nodes = self.apitable.space(SPACE_ID).nodes.all()
        self.assertIsInstance(space_root_nodes, list)
        first_node = space_root_nodes[0]
        self.assertIn(first_node.type, ["Datasheet", "Folder"])

    def test_get_node_detail(self):
        node = self.apitable.nodes.get(FOLDER_ID)
        self.assertEqual(node.id, FOLDER_ID)
        self.assertIsNotNone(node.children)


if __name__ == "__main__":
    unittest.main()



import time
import requests


class ApitableSession(requests.Session):
    def __init__(self):
        super(ApitableSession, self).__init__()
        self.qps = 1 / 5
        # self.qps_map = dict()

    def get(self, *args):
        if self.qps > 0:
            time.sleep(self.qps)
        return super(ApitableSession, self).get(*args)

    def post(self, *args):
        if self.qps > 0:
            time.sleep(self.qps)
        return super(ApitableSession, self).post(*args)

    def patch(self, *args):
        if self.qps > 0:
            time.sleep(self.qps)
        return super(ApitableSession, self).patch(*args)

    def delete(self, *args):
        if self.qps > 0:
            time.sleep(self.qps)
        return super(ApitableSession, self).delete(*args)



import unittest
import warnings

from vika import Vika
from . import TOKEN, DOMAIN, SPACE_ID, DATASHEET_ID, VIEW_ID


class TestGetRecords(unittest.TestCase):
    def setUp(self):
        warnings.simplefilter("ignore", ResourceWarning)
        apitable = Vika(TOKEN)
        apitable.set_api_base(DOMAIN)
        self.dst = apitable.space(SPACE_ID).datasheet(DATASHEET_ID)

    def test_record_count(self):
        self.assertEqual(self.dst.records.all().count(), 1)

    def test_record_filter_get(self):
        self.assertEqual(
            self.dst.records.filter(title="apitable").get().title, "apitable"
        )

    def test_record_all(self):
        self.assertEqual(
            self.dst.records.all(viewId=VIEW_ID).count(), 1)

    def test_get_or_create(self):
        # Ger record
        record, created = self.dst.records.get_or_create(title="apitable")
        self.assertFalse(created)

        # Create record
        record, created = self.dst.records.get_or_create(title="nonexistent record")
        record.delete()
        self.assertEqual(record.title, "nonexistent record")
        self.assertTrue(created)

    # FXIME: Now there is a problem with the return of the rest api, so don't add this first
    # def test_record_all_with_params(self):
    #     # Nonexistent pagination returns empty records
    #     self.assertEqual(self.dst.records.all(pageNum=2).count(), 0)


if __name__ == "__main__":
    unittest.main()



import os
import time
import unittest
import warnings

from vika import Vika

from . import TOKEN, DOMAIN, SPACE_ID, DATASHEET_ID


class TestUploadFile(unittest.TestCase):
    def setUp(self):
        warnings.simplefilter("ignore", ResourceWarning)
        apitable = Vika(TOKEN)
        apitable.set_api_base(DOMAIN)
        self.dst = apitable.space(SPACE_ID).datasheet(DATASHEET_ID)

    def test_upload_file(self):
        test_url = "https://i0.wp.com/apitable.com/wp-content/uploads/2022/11/IT.png"
        test_file = self.dst.upload_attachment(test_url)
        self.assertIsNotNone(test_file.get("token"))

        time.sleep(2)

        test_local_file = "test.png"
        filepath = os.path.join(os.path.dirname(__file__), test_local_file)
        test_file = self.dst.upload_file(filepath)
        self.assertEqual(test_file.get("mimeType"), "image/png")


if __name__ == "__main__":
    unittest.main()
